import os
from datetime import datetime

import numpy as np
import pandas as pd


class AgeFromName(object):
	def __init__(self, mortality_df=None, year_of_birth_df=None):
		'''
		:param mortality_df: pd.DataFrame, optional
		:param year_of_birth_df: pd.DataFrame, optional
		'''
		if mortality_df is None:
			self._mortality_df = pd.read_csv(self._get_data_path('mortality_table.csv.gz'))
		else:
			self._mortality_df = mortality_df
		if year_of_birth_df is None:
			self._year_of_birth_df = pd.read_csv(self._get_data_path('year_of_birth_counts.csv.gz'))
		else:
			self._year_of_birth_df = year_of_birth_df

	def _get_data_path(self, file_name):
		return os.path.join(os.path.dirname(__file__), 'data', file_name)

	def get_estimated_counts(self, first_name, sex, current_year=datetime.now().year):
		'''
		:param first_name: str, First name
		:param sex: str, m or f for sex
		:param current_year: int, optional, defaults to current year
		:return: pd.Series, with int indices indicating years of
			birth, and estimated counts of total population with that name and birth year
		'''
		first_name = first_name.lower()
		sex = sex.lower()
		cur_df = (self._year_of_birth_df[(self._year_of_birth_df.first_name == first_name)
		                                 & (self._year_of_birth_df.sex == sex)
		                                 & (self._year_of_birth_df.year_of_birth <= current_year)]
		          [['year_of_birth', 'count']])
		year_stats = (self._mortality_df[self._mortality_df.as_of_year == current_year]
		              [['year_of_birth', sex + '_prob_alive']])
		cur_df['prob_alive'] = np.interp(cur_df.year_of_birth,
		                                 year_stats.year_of_birth,
		                                 year_stats[sex + '_prob_alive'])
		cur_df['estimated_count'] = cur_df['prob_alive'] * cur_df['count']
		return cur_df.set_index('year_of_birth')['estimated_count']

	def get_mle(self, first_name, sex, current_year=datetime.now().year):
		'''
		:param first_name: str, First name
		:param sex: str, m or f for sex
		:param current_year: int, optional, defaults to current year
		:return: int, the most likely year of birth
		'''
		return self.get_estimated_counts(first_name, sex, current_year).argmax()

	def get_estimated_distribution(self, first_name, sex, current_year=datetime.now().year):
		'''
		:param first_name: str, First name
		:param sex: str, m or f for sex
		:param current_year: int, optional, defaults to current year
		:return: pd.Series, with int indices indicating years of
			birth, and the estimated percentage of the total population of people who share sex and
			first name who were born that year.
		'''
		age_counts = self.get_estimated_counts(first_name, sex, current_year)
		to_ret = age_counts / age_counts.sum()
		to_ret.name = 'estimate_percentage'
		return to_ret



import datetime
import io
import math
from urllib.request import urlopen
from zipfile import ZipFile

import numpy as np
import pandas as pd
from bs4 import BeautifulSoup


def regenerate_birth_counts(
		census_zip_file=None,
		output_path='data/year_of_birth_counts.csv.gz'):
	'''Regenerate table containing counts of first names by sex and year of birth.

	:param census_zip_file: str, file-like object similar to http://www.ssa.gov/oact/babynames/names.zip, defaults to SSA.gov url
	:param output_path: str, path of .gz file to write dataframe csv
	:return: pd.DataFrame, pandas data frame with the columns first_name,sex,count,year_of_birth
	'''
	if census_zip_file is None:
		census_zip_file = io.BytesIO(urlopen('http://www.ssa.gov/oact/babynames/names.zip').read())
	year_of_birth_dfs = []
	with ZipFile(census_zip_file) as names_zip:
		for filename in names_zip.namelist():
			if filename.startswith('yob') and filename.endswith('txt'):
				cur_year_of_birth_df = pd.read_csv(names_zip.open(filename),
				                                   index_col=None,
				                                   names=['first_name', 'sex', 'count'])
				cur_year_of_birth_df['year_of_birth'] = filename[3:7]
				year_of_birth_dfs.append(cur_year_of_birth_df)
	year_of_birth_df = pd.concat(year_of_birth_dfs, ignore_index=True)
	year_of_birth_df['first_name'] = year_of_birth_df['first_name'].apply(str.lower)
	year_of_birth_df['sex'] = year_of_birth_df['sex'].apply(str.lower)
	year_of_birth_df.to_csv(output_path, index=False, compression='gzip')
	return year_of_birth_df


def _decade_mortality_table(year,
                            url_template='https://www.ssa.gov/oact/NOTES/as120/LifeTables_Tbl_7_{}.html'):
	assert int(year) % 10 == 0
	url = url_template.format(year)
	soup = BeautifulSoup(urlopen(url).read(), 'lxml')
	table = soup.find('table', border=1)
	rows = []
	for row in table.find_all('tr'):
		row_datum = [cell.text.strip() for cell in row.find_all('td')]
		if len(row_datum) == 15 and row_datum[0] != '':
			rows.append({
				'year_of_birth': int(year),
				'age': int(row_datum[0]),
				'm_prob_survive_that_year': 1 - float(row_datum[1]),
				'f_prob_survive_that_year': 1 - float(row_datum[9]),
			})
	df = pd.DataFrame(rows).sort_values(by='age')
	for sex in 'mf':
		df[sex + '_prob_alive'] = np.cumprod(df[sex + '_prob_survive_that_year']).astype(np.float64)
	df['as_of_year'] = df['year_of_birth'] + df['age']
	return df[['year_of_birth', 'as_of_year', 'm_prob_alive', 'f_prob_alive']]


def regenerate_decade_mortality_table(
		url_template='https://www.ssa.gov/oact/NOTES/as120/LifeTables_Tbl_7_{}.html',
		output_path='data/mortality_table.csv.gz',
		min_decade=1900,
		max_decade=math.ceil(datetime.datetime.now().year * 0.1) * 10):
	'''
	:param url_template: str, url tempate (with year as {}) to scrape
	:param output_path: str, path of .gz file to write dataframe csv
	:param min_decade: int, minimum decade to search
	:param max_decade: int, maximum decade to search
	:return: pd.DataFrame, pandas data frame with the columns year_of_birth,as_of_year,m_death_prob,f_death_prob
	'''
	mortality_df = pd.concat([_decade_mortality_table(year, url_template)
	                          for year
	                          in range(min_decade, max_decade, 10)], axis=0)
	mortality_df.to_csv(output_path, index=False, compression='gzip')
	return mortality_df


def regenerate_all():
	regenerate_birth_counts()
	regenerate_decade_mortality_table()



from datetime import datetime
import pandas as pd
from agefromname import AgeFromName


class InvalidGenerationBirthYearDefinition(Exception):
	pass


class GenerationFromName(object):
	def __init__(self,
	             generation_birth_years={'Greatest Generation': [1930, 1945],
	                                     'Baby Boomers': [1946, 1964],
	                                     'Generation X': [1965, 1980],
	                                     'Millenials': [1981, 2000],
	                                     'Generation Z': [2001, 2020]},
	             age_from_name=None):
		'''
		:param generation_birth_years: dict, maps generation names to the first and
			last years of birth. Ex.: {'Millenials': [1980, 1995], 'Generation X': [1956, 1979]}
			Years outside this range will be considered "_other".
		:param age_from_name: AgeFromName, optional.  If note entered, will be autogenerated
		'''
		self._validate_generation_birth_years(generation_birth_years)
		self._generation_birth_years = generation_birth_years
		self._age_from_name = age_from_name if age_from_name is not None else AgeFromName()

	def _validate_generation_birth_years(self, generation_birth_years):
		invalid_type_or_tempate_error = "generation_birth_years must be a dict, which maps generation names to first and last birth years.  Ex: {'Millenials': [1980, 1995],'Generation X': [1956, 1979]}."
		if type(generation_birth_years) != dict:
			raise InvalidGenerationBirthYearDefinition(invalid_type_or_tempate_error)
		for key, val in generation_birth_years.items():
			if type(val) not in (list, tuple):
				raise InvalidGenerationBirthYearDefinition("Values in generation_birth_years must be lists or tuples.")
			if len(val) != 2:
				raise InvalidGenerationBirthYearDefinition(
					"Values in generation_birth_years must be nondecreasingly ordered lists of length two.")
			if val[0] > val[1]:
				raise InvalidGenerationBirthYearDefinition(
					"Values in generation_birth_years must be nondecreasingly ordered lists of length two.")

		ordered_ranges = list(sorted(generation_birth_years.values()))
		for i in range(1, len(ordered_ranges)):
			if ordered_ranges[i - 1][1] >= ordered_ranges[i][0]:
				raise InvalidGenerationBirthYearDefinition(
					"Values in generation_birth_years non overlapping. %s and %s overlap." % (
					ordered_ranges[i - 1], ordered_ranges[i]))

	def get_estimated_counts(self, first_name, sex, current_year=datetime.now().year):
		'''
		:param first_name: str, First name
		:param sex: str, m or f for sex
		:param current_year: int, optional, defaults to current year
		:return: pd.Series, with int indices indicating years of
			birth, and estimated counts of total population with that name and generation
		'''
		year_counts = self._age_from_name.get_estimated_counts(first_name, sex, current_year)
		to_ret = self._generational_rollup(year_counts)
		to_ret.name = 'estimated_count'
		return to_ret

	def get_estimated_distribution(self, first_name, sex, current_year=datetime.now().year):
		'''
		:param first_name: str, First name
		:param sex: str, m or f for sex
		:param current_year: int, optional, defaults to current year
		:return: pd.Series, nd the estimated percentage of the total population of
		people who share sex andfirst name who were born that generation.
		'''
		to_ret = self._generational_rollup(self._age_from_name.get_estimated_distribution(first_name, sex, current_year))
		to_ret.name = 'estimate_percentage'
		return to_ret

	def get_mle(self, first_name, sex, current_year=datetime.now().year):
		'''
		:param first_name: str, First name
		:param sex: str, m or f for sex
		:param current_year: int, optional, defaults to current year
		:return: pd.Series, nd the estimated percentage of the total population of
		people who share sex andfirst name who were born that generation.
		'''
		return self.get_estimated_distribution(first_name, sex, current_year).argmax()


	def _generational_rollup(self, year_counts):
		generation_counts = {generation: year_counts[(year_counts.index <= genmax)
		                                             & (year_counts.index >= genmin)].sum()
		                     for generation, (genmin, genmax)
		                     in self._generation_birth_years.items()}
		generation_counts['_other'] = year_counts.sum() - sum(generation_counts.values())
		to_ret = pd.Series(generation_counts)
		return to_ret





from collections import Counter
from unittest import TestCase
import pandas as pd
from agefromname import AgeFromName
from agefromname.generation_from_name import GenerationFromName, InvalidGenerationBirthYearDefinition


class TestGenerationFromName(TestCase):
	def test_validation(self):
		g = GenerationFromName()
		a = AgeFromName()
		with self.assertRaises(InvalidGenerationBirthYearDefinition):
			g = GenerationFromName(3, a)
		with self.assertRaises(InvalidGenerationBirthYearDefinition):
			g = GenerationFromName({'AAA': {3, 1}}, a)
		with self.assertRaises(InvalidGenerationBirthYearDefinition):
			g = GenerationFromName({'AAA': [3, 1, 1]}, a)
		with self.assertRaises(InvalidGenerationBirthYearDefinition):
			g = GenerationFromName({'AAA': [1]}, a)
		g = GenerationFromName({'AAA': [1999, 1999]}, a)
		with self.assertRaises(InvalidGenerationBirthYearDefinition):
			g = GenerationFromName({'AAA': [1999, 1888]}, a)
		with self.assertRaises(InvalidGenerationBirthYearDefinition):
			g = GenerationFromName({'AAA': [1999, 1888]}, a)
		with self.assertRaises(InvalidGenerationBirthYearDefinition):
			g = GenerationFromName({'AAA': [1999, 1999], 'BBB': [1888, 2000]}, a)
		with self.assertRaises(InvalidGenerationBirthYearDefinition):
			g = GenerationFromName({'AAA': [1990, 2000], 'BBB': [1985, 1990]}, a)
		with self.assertRaises(InvalidGenerationBirthYearDefinition):
			g = GenerationFromName({'AAA': [1990, 2000], 'BBB': [1985, 1990]}, a)
		g = GenerationFromName({'AAA': [1990, 2000], 'BBB': [2001, 2010]}, a)

	def test_get_estimated_counts(self):
		g = GenerationFromName({'AAA': [1930, 1945], 'BBB': [1946, 1999]})
		year_count = AgeFromName().get_estimated_counts('nancy', 'f', 1950)
		actual = g.get_estimated_counts('nancy', 'f', 1950)
		self.assertEqual(type(actual), pd.Series)
		self.assertEqual(actual.name, 'estimated_count')
		self._manually_verify_counts_or_distribution(actual, year_count)

		self.assertAlmostEqual(sum(actual), sum(year_count), places=0)


	def test_get_estimated_distribution(self):
		g = GenerationFromName({'AAA': [1930, 1945], 'BBB': [1946, 1999]})
		year_distribution = AgeFromName().get_estimated_distribution('nancy', 'f', 1950)
		actual = g.get_estimated_distribution('nancy', 'f', 1950)
		self.assertEqual(type(actual), pd.Series)
		self.assertEqual(actual.name, 'estimate_percentage')
		self._manually_verify_counts_or_distribution(actual, year_distribution)

		self.assertAlmostEqual(sum(actual), 1, places=0)

	def _manually_verify_counts_or_distribution(self, actual, year_count):
		self.assertEqual(set(actual.keys()),
		                 {'AAA', 'BBB', '_other'})
		expected = Counter()
		for year, cnt in year_count.iteritems():
			if 1930 <= year <= 1945:
				expected['AAA'] += cnt
			elif 1946 <= year <= 1999:
				expected['BBB'] += cnt
			else:
				expected['_other'] += cnt
		self.assertAlmostEqual(expected['AAA'], actual['AAA'], places=0)
		self.assertAlmostEqual(expected['BBB'], actual['BBB'], places=0)
		self.assertAlmostEqual(expected['_other'], actual['_other'], places=0)



from setuptools import setup, find_packages

setup(name='agefromname',
      version='0.0.1.1',
      description='Predict how old someone is from their name and gender, using the baby name data from the US Social Security Administration.',
      url='https://github.com/JasonKessler/agefromname',
      author='Jason Kessler',
      author_email='jason.kessler@gmail.com',
      license='MIT',
      packages=find_packages(),
      install_requires=[
	      'nose',
	      'numpy',
	      'pandas',
	      'beautifulsoup4'
      ],
      package_data={
	      'agefromname': ['data/*']
      },
      test_suite="nose.collector",
      tests_require=['nose'],
      setup_requires=['nose>=1.0'],
      zip_safe=False)



from unittest import TestCase
import pandas as pd
from agefromname.age_from_name import AgeFromName


class TestBirthYearPredictor(TestCase):
	def setUp(cls):
		cls.birth_year_predictor = AgeFromName()

	def test_get_mle(self):
		birth_year_predictor = self.birth_year_predictor
		self.assertEqual(birth_year_predictor.get_mle('jason', 'm', 2017), 1977)
		self.assertEqual(birth_year_predictor.get_mle('JAsOn', 'M', 2017), 1977)
		self.assertEqual(birth_year_predictor.get_mle('jennifer', 'f', 2017), 1972)
		self.assertEqual(birth_year_predictor.get_mle('jeNNifer', 'F', 2017), 1972)
		self.assertEqual(birth_year_predictor.get_mle('nancy', 'f', 2017), 1952)

	def test_get_mle_default(self):
		self.birth_year_predictor.get_mle('nancy', 'f')

	def test_mle_not_current_year(self):
		self.assertEqual(self.birth_year_predictor.get_mle('nancy', 'f', 1901), 1900)

	def test_get_estimated_distribution(self):
		actual = self.birth_year_predictor.get_estimated_distribution('nancy', 'f', 1901)
		self.assertEqual(type(actual), pd.Series)
		self.assertEqual(actual.name, 'estimate_percentage')
		self.assertEqual(set(actual.keys()),
		                 {1880, 1881, 1882, 1883, 1884, 1885, 1886, 1887, 1888, 1889,
		                  1890, 1891, 1892, 1893, 1894, 1895, 1896, 1897, 1898, 1899,
		                  1900, 1901})
		self.assertAlmostEqual(sum(actual), 1)

	def test_get_estimated_counts(self):
		actual = self.birth_year_predictor.get_estimated_counts('nancy', 'f', 1901)
		self.assertEqual(type(actual), pd.Series)
		self.assertEqual(actual.name, 'estimated_count')
		self.assertEqual(set(actual.keys()),
		                 {1880, 1881, 1882, 1883, 1884, 1885, 1886, 1887, 1888, 1889,
		                  1890, 1891, 1892, 1893, 1894, 1895, 1896, 1897, 1898, 1899,
		                  1900, 1901})
		self.assertAlmostEqual(sum(actual), 9669, places=0)